#include "pendulum_parametric.h"
using namespace std;

pendulum_parametric::pendulum_parametric() : epsilon(1), delta(1), L(1), g(9.81), M(1), X0({0.1, 0}) { Init(); } //Standard Constructor

pendulum_parametric::pendulum_parametric(double d, double e, double l, double G, double m, const Xvar &x) : epsilon(e), delta(d), L(l), g(G), M(m), X0(x) { Init(); } //Constructor that takes values for L, g, m and initial conditions X0 other than the asked

pendulum_parametric::pendulum_parametric(double d, double e, const Xvar &x) : epsilon(e), delta(d), L(1), g(9.81), M(1), X0(x) { Init(); } //XVar constructor

pendulum_parametric::pendulum_parametric(double d, double e, const std::initializer_list<double> &v) : epsilon(e), delta(d), L(1), g(9.81), M(1), X0(v) { Init(); } //Initializer list constructor

pendulum_parametric::pendulum_parametric(double d, double e, double theta_0, double theta_vel_0) : epsilon(e), delta(d), L(1), g(9.81), M(1), X0({theta_0, theta_vel_0}) { Init(); } //Initial Conditions as Doubles

pendulum_parametric::~pendulum_parametric() //Class destructor
{
   DimSol.clear();
   AdimSol.clear();
   AverageSol.clear();
}

//Create the lambda functions to solve the differential equation
void pendulum_parametric::Init()
{
   f[0] = [](ODEpoint &x)
   { return x[1]; };
   f[1] = [this](ODEpoint &x)
   { return -sin(x[0]) * (this->delta + (this->epsilon) * cos(x.T())); };
}

//RK4 method
void pendulum_parametric::RungeKutta4(double step, double Time)
{
   //Variables for the Nondimensionalization
   double fatort = sqrt((L * delta) / g);
   double fatorw = 1 / fatort;
   int n = floor(Time / step);
   double step2 = step / fatort;
   //Temporary vector of ODEpoints to store the solution
   vector<ODEpoint> solutionAdim;
   vector<ODEpoint> solutionDim;
   solutionAdim.resize(n + 1);
   solutionDim.resize(n + 1);
   //Temporary vector of vectors to store the values of theta and omega, initializing it with 0
   vector<vector<double>> vec;
   for (int j = 0; j < n + 1; ++j)
   {
      vec.push_back({0, 0});
   }
   //Initial conditions for the Nondimensionalized conditions
   vec[0][0] = X0[0];
   vec[0][1] = X0[1] / fatorw;

   //Store the Matrix of K values
   double K[4][2];
   //Auxiliar variables
   double iter;
   int j;
   for (int i = 0; i < n; ++i)
   {
      solutionAdim[i].SetODEpoint(i * step2, {vec[i][0], vec[i][1]}); //Temporary solution

      for (j = 0; j < 3; j++) //Calculates K11, K12, K21, K22, K31, K32
      {
         iter = floor((j + 2) / 2);
         K[j][0] = f[0](solutionAdim[i]);
         K[j][1] = f[1](solutionAdim[i]);
         //Use K to calculate the intermedeary next iteration of vec
         vec[i + 1][0] = vec[i][0] + step2 * K[j][0] * iter / 2;
         vec[i + 1][1] = vec[i][1] + step2 * K[j][1] * iter / 2;
         solutionAdim[i].SetODEpoint((i + iter / 2) * step2, {vec[i + 1][0], vec[i + 1][1]}); //Temporary solution
      }
      K[3][0] = f[0](solutionAdim[i]); //Calculate K41
      K[3][1] = f[1](solutionAdim[i]); //Calculate K42
      //Calculate the final value for the next iteration of vec
      vec[i + 1][0] = vec[i][0] + step2 / 6 * (K[0][0] + 2 * K[1][0] + 2 * K[2][0] + K[3][0]);
      vec[i + 1][1] = vec[i][1] + step2 / 6 * (K[0][1] + 2 * K[1][1] + 2 * K[2][1] + K[3][1]);
      solutionAdim[i].SetODEpoint(i * step2, {vec[i][0], vec[i][1]});                  //Store the final nondimensionalized solution
      solutionDim[i].SetODEpoint(i * step2 * fatort, {vec[i][0], fatorw * vec[i][1]}); //Store the final dimensionalized solution
   }
   solutionAdim[n].SetODEpoint(n * step2, {vec[n][0], vec[n][1]});                  //Store the last adimensionalized solution
   solutionDim[n].SetODEpoint(n * step2 * fatort, {vec[n][0], fatorw * vec[n][1]}); //Store the last dimensionalized solution

   DimSol = solutionDim;
   AdimSol = solutionAdim;
   //Clean
   solutionAdim.clear();
   solutionDim.clear();
   vec.clear();
}

//Check if the moving average of a given soluton constitutes a stable solution or not,
//depending on whether its module is greater than pi
bool pendulum_parametric::CheckStability()
{
   int size = AverageSol.size();
   for (int i = 0; i < size; i++)
   {
      if (AverageSol[i][0] > M_PI)
      {
         return 0;
      }
   }
   return 1;
}

//Getter for the dimensional solution dimension
std::vector<ODEpoint> &pendulum_parametric::GetDimSolution()
{
   return DimSol;
}

//Getter for the nondimensional solution dimension
std::vector<ODEpoint> &pendulum_parametric::GetAdimSolution()
{
   return AdimSol;
}

//Draw the solution (if vardim=1, draw dimensionalized graph, if vardim=0, draw nondimensionalized)
void pendulum_parametric::Draw(bool vardim, double tf)
{  
   //Create c arrays for ploting
   double *Theta = new double[DimSol.size()];
   double *Time = new double[DimSol.size()];
   for (int i = 0; i < DimSol.size(); ++i)
   {
      if (vardim)
      {
         Theta[i] = DimSol[i][0]; //Store values of theta and time for dimensional solution
         Time[i] = DimSol[i].T();
      }
      else
      {
         Theta[i] = AdimSol[i][0]; //Store values of theta hat and time hat for adimensional solution
         Time[i] = AdimSol[i].T();
         tf = (AdimSol.back()).T();
      }
   }

   TCanvas *c1 = new TCanvas("c1", "Pendulum", 100, 100, 900, 500); //Crate canvas and graph
   TGraph *gr1 = new TGraph(DimSol.size(), Time, Theta);

   if (vardim) //Graph for the dimensional solution
   {
      gr1->SetMarkerColor(4); //Settings for graph
      gr1->SetLineColor(4);
      gr1->SetTitle("Grafico de #theta(t);t(segundos);#theta(radianos)");
   }

   else //Graph for the nondimensional solution
   {
      gr1->SetMarkerColor(2); //Settings for graph
      gr1->SetLineColor(2);
      gr1->SetTitle("Grafico de #hat{#theta}(#hat{t});#scale[1.3]{#hat{t}};#scale[1.3]{#hat{#theta}}");
   }

   gr1->GetXaxis()->SetLimits(0, tf); //Set axis and labels
   gr1->GetYaxis()->SetTitleOffset(0.8);
   gr1->Draw("ALP");
   c1->Modified();
   c1->Update();

   if (vardim) //Save the plot
      c1->SaveAs("ThetaVsT.png");
   else
      c1->SaveAs("ThetaHatVsT.png");

   c1->Clear(); //Clear memory
   delete c1, gr1;
   delete[] Theta, Time;
}

//Draw the solutions' moving average graph (which is always nondimensional)
void pendulum_parametric::DrawAverageGraph()
{
   double *ThetaAve = new double[AverageSol.size()]; //Arrays for storing and plotting
   double *TimeAve = new double[AverageSol.size()];

   for (int i = 0; i < AverageSol.size(); ++i)
   {
      ThetaAve[i] = AverageSol[i][0]; //Store average values
      TimeAve[i] = AverageSol[i].T();
   }

   double *Theta = new double[AdimSol.size()]; //Arrays for storing adimensional values
   double *Time = new double[AdimSol.size()];

   for (int i = 0; i < AdimSol.size(); ++i)
   {

      Theta[i] = fabs(AdimSol[i][0]); //Storing
      Time[i] = AdimSol[i].T();
   }

   TCanvas *c1 = new TCanvas("c1", "Pendulum", 100, 100, 900, 500); //Canvas and graph for plotting
   TGraph *gr2 = new TGraph(AdimSol.size(), Time, Theta);
   ostringstream streamdelta; //Stringstreams to make the title variable with delta and epsilon
   streamdelta << delta;
   string strdelta = streamdelta.str();
   ostringstream streamepsilon;
   streamepsilon << epsilon;
   string strepsilon = streamepsilon.str();
   string SaveFile = "GraficoDaMediaDeslizante";
   SaveFile = SaveFile + "Delta_" + strdelta + "Epsilon_" + strepsilon + ".png"; //Name of the save file
   string Title = "Grafico da media deslizante sobreposto com | #hat{#theta}(#hat{t}) | para #delta=";
   Title = Title + strdelta + " e #varepsilon=" + strepsilon + ";#scale[1.3]{#hat{t}};#scale[1.3]{#hat{#theta} // <#hat{#theta}(#hat{t})>}"; //Title and name of axis

   gr2->SetMarkerColor(4); //Graph settings 
   gr2->SetLineColor(4);
   gr2->GetXaxis()->SetLimits(0, AdimSol.back().T());
   gr2->SetTitle(Title.c_str());
   gr2->Draw("ALP");
   c1->Modified();
   c1->Update();
   TGraph *gr1 = new TGraph(AverageSol.size(), TimeAve, ThetaAve);
   TLegend *label = new TLegend(0.1, 0.8, 0.25, 0.9);      //Create a legend
   label->SetHeader("Legenda", "C");                       //Create title for the legend
   label->AddEntry(gr1, "< #hat{#theta}(#hat{t}) >", "l"); //Add entries to the legend
   label->AddEntry(gr2, " | #hat{#theta}(#hat{t}) |", "l");
   gr1->SetMarkerColor(2);
   gr1->SetLineColor(2);
   gr1->Draw("Same"); //Draw graphs
   label->Draw();
   c1->Modified();
   c1->Update();
   c1->SaveAs(SaveFile.c_str()); //Save plotting
   c1->Clear();

   delete c1, gr1, gr2, label;
   delete[] Theta, Time, ThetaAve, TimeAve; //Clear memory 
}

//Draw the phase diagram
void pendulum_parametric::DrawPhaseDiagram(bool vardim)
{
   double *Theta = new double[DimSol.size()]; //Arrays of values for plotting
   double *Omega = new double[DimSol.size()];
   for (int i = 0; i < DimSol.size(); ++i)
   {
      //Phase diagram for the dimensional solution
      if (vardim)
      {
         Theta[i] = DimSol[i][0]; //Store dimensional values
         Omega[i] = DimSol[i][1];
      }
      //Phase diagram for the nondimensional solution
      else
      {
         Theta[i] = AdimSol[i][0]; //Store adimensional values
         Omega[i] = AdimSol[i][1];
      }
   }
   TGraph *gr1 = new TGraph(DimSol.size(), Theta, Omega); //Graph and canvas for plotting
   TCanvas *c1 = new TCanvas("c1", "Pendulum", 100, 100, 900, 500);
   ostringstream streamdelta; //Stringstreams for titles with variables delta and epsilon 
   streamdelta << delta;
   string strdelta = streamdelta.str();
   ostringstream streamepsilon;
   streamepsilon << epsilon;
   string strepsilon = streamepsilon.str();
   string SaveFile = "DiagramaDeFase";
   SaveFile = SaveFile + "Delta_" + strdelta + "Epsilon_" + strepsilon;
   string Title = "Diagrama de fase";
   //Titles according to whether the solution is dimensional or nondimensional
   if (vardim)
   {
      Title = Title + " dimensionalizado com #delta= " + strdelta + " e #varepsilon= " + strepsilon + ";#theta(radianos);#omega(radianos/s)";
      gr1->SetMarkerColor(4);
      gr1->SetLineColor(4);
      gr1->SetTitle(Title.c_str());
   }
   else
   {
      Title = Title + " adimensionalizado com #delta= " + strdelta + " e #varepsilon= " + strepsilon + ";#scale[1.3]{#hat{#theta}};#scale[1.3]{#hat{#omega}}";
      gr1->SetMarkerColor(2);
      gr1->SetLineColor(2);
      gr1->SetTitle(Title.c_str());
   }
   gr1->Draw("ALP");//Draw graph
   c1->Modified();
   c1->Update();
   //File name according to whether the solution is dimensional or nondimensional
   if (vardim)
   {
      SaveFile = SaveFile + "Dim.png";
      c1->SaveAs(SaveFile.c_str());
   }
   else
   {
      SaveFile = SaveFile + "Adim.png";
      c1->SaveAs(SaveFile.c_str());
   }
   delete [] Theta, Omega;
   delete gr1, c1;
}

//Calculating the moving average of a given solution with respect to a given tw (argument)
void pendulum_parametric::MovingAverage(double tw)
{
   double step = (AdimSol[1].T() - AdimSol[0].T());
   double tf = (AdimSol.back()).T();
   int i;
   double sum = 0;
   AverageSol.resize(AdimSol.size());
   
   //Total_Time=(AdimSol.size()-1)*step;
   //Calculating from 0 to tw/(2*step) (asymmetrical interval), which gives the moving average in 0
   for (i = 0; i <= (tw / 2) / step; ++i)
   {
      sum += (AdimSol[i][0] * AdimSol[i][0]);
   }

   AverageSol[0].SetODEpoint(0, {sqrt(sum / i)});
   i--;
   
   //Calculating from tw/2 to tw (asymmetrical interval),
   //which gives the moving average in points from time (step) to (tw/2+step)
   for (int j = 1; j <= (tw / 2) / step; ++j)
   {
      sum += (AdimSol[i + j][0] * AdimSol[i + j][0]);
      AverageSol[j].SetODEpoint(j * step, {sqrt(sum / (i + j + 1))});
   }

   int n = 2 * i + 1;
   int k;

   //Calculating from tw to AdimSol.size()-1-tw/step (symmetrical interval), 
   //which gives the moving average in points from time (tw/2+2*step) to (Total_Time - tw/2)
   for (k = 0; k < (AdimSol.size() - n - 1); ++k)
   {
      sum += AdimSol[n + k][0] * AdimSol[n + k][0];
      sum -= AdimSol[k][0] * AdimSol[k][0];
      AverageSol[i + k + 1].SetODEpoint((i + k + 1) * step, {sqrt(sum / n)});
   }

   //Calculating from AdimSol.size()-tw/step to AdimSol.size()-1 (asymmetrical interval),
   //which gives the moving average in points from (Total_Time - tw/2+1) to time (Total_Time)
   for (int j = 0; j < (tw / 2) / step; ++j) 
   {
      sum -= AdimSol[k + j][0] * AdimSol[k + j][0];
      AverageSol[i + j + k + 1].SetODEpoint((i + j + k + 1) * step, {sqrt(sum / (n - j - 1))});
   }
}