#include "ODESolver.h"
#include <sstream>
#include "TPDF.h"
#include "Roots.h"
using namespace std;

ODESolver::ODESolver() : Ndim(1), X0({1}), onequation(1)
{
    auto f1 = [](ODEpoint &x)
    { return x[0]; };
    f.push_back(f1);
    factor.push_back(1);
    factort = 1;
} //Standard Constructor

ODESolver::ODESolver(const int m, const Xvar &x, vector<function<double(ODEpoint &)>> &System, std::vector<double> &initfactor, double time, const bool a) : Ndim(m), X0(x), f(System), onequation(a), factor(initfactor), factort(time) { ; } //XVar constructor

ODESolver::ODESolver(const int m, const Xvar &x, vector<function<double(ODEpoint &)>> &System, const bool a) : Ndim(m), X0(x), f(System), onequation(a) { ; } //XVar constructor

ODESolver::ODESolver(const int m, const std::initializer_list<double> &v, vector<function<double(ODEpoint &)>> &System, std::vector<double> &initfactor, double time, const bool a) : Ndim(m), X0(v), f(System), onequation(a), factor(initfactor), factort(time) { ; } //Initializer list constructor

ODESolver::ODESolver(const int m, const std::initializer_list<double> &v, vector<function<double(ODEpoint &)>> &System, const bool a) : Ndim(m), X0(v), f(System), onequation(a) { ; } //Initializer list constructor

ODESolver::~ODESolver(){
    MS.clear();
    MDS.clear();
    f.clear();
    factor.clear(); 
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

const std::vector<ODEpoint> &ODESolver::VerletSolver(double step, double Time) //Using the Verlet Method
{
    if (Ndim == 2)
    {
        int n = floor(Time / (step));

        //Temporary vector of ODEpoints to store the solution
        vector<ODEpoint> solution;
        solution.resize(n + 1);

        //Temporary vector of vectors to store the values of theta and omega, initializing it with 0
        vector<vector<double>> vec;
        for (int j = 0; j < n + 1; ++j)
        {
            vec.push_back({0, 0});
        }

        //Initial conditions for the Nondimensionalized conditions
        vec[0][0] = X0[0];
        vec[0][1] = X0[1];
        solution[0].SetODEpoint(0, {X0[0], X0[1]});                                     //Store it at the solution vector
        vec[1][0] = X0[0] + (step * X0[1]) + ((step * step) / 2) * (f[1](solution[0])); //Calculate the first iteration for theta

        for (int i = 1; i < n; ++i)
        {
            solution[i].SetODEpoint(i * step, {vec[i][0], vec[i][1]});                          //Temporarily create a solution to use in the next calculations
            vec[i + 1][0] = 2 * vec[i][0] - vec[i - 1][0] + pow(step, 2) * (f[1](solution[i])); //Calculate the next theta
            vec[i][1] = (vec[i + 1][0] - vec[i - 1][0]) / (2 * step);                           //Calculate omega
            solution[i].SetODEpoint(i * step, {vec[i][0], vec[i][1]});                          //Store the real solution, considering the nondimensionalization
        }

        double last_theta = 2 * vec[n][0] - vec[n - 1][0] + step * step * (-sin(vec[n][0])); //Calculate an extra theta to get the last omega
        vec[n][1] = (last_theta - vec[n - 1][0]) / (2 * step);                               //Last omega
        solution[n].SetODEpoint(n * step, {vec[n][0], vec[n][1]});                           //Store last solution, considering the nondimensionalization

        MS.insert({"Verlet", solution}); //Store the solution
        //Clean
        solution.clear();
        vec.clear();
    }
    else
    {
        cout << "The Stormer-Verlet Method is only for two dimensional differential problems, the dimension inserted is not valid, the equivelent method for n dimensions is the leap-frog method" << endl;
        ODEpoint R(0, X0);
        vector<ODEpoint> Temp;
        Temp.push_back(R);
        MS.insert({"Verlet", Temp});
        Temp.clear();
    }
    return MS.find("Verlet")->second; //Return the solution
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

const std::vector<ODEpoint> &ODESolver::LeapFrog(double step, double Time)
{
    cout << "Entrou" << endl;
    int n = floor(Time / step);
    vector<ODEpoint> solution;
    solution.resize(n + 1);
    vector<vector<double>> vec;
    vec.resize(n + 1);
    for (int j = 0; j < n + 1; ++j)
    {
        vec[j].resize(Ndim);
    }

    for (int j = 0; j < Ndim; ++j)
    {
        vec[0][j] = X0[j];
    }

    solution[0].SetODEpoint(0, vec[0]);

    for (int j = 0; j < Ndim; ++j)
    {
        vec[1][j] = vec[0][j] + step * f[j](solution[0]);
    }
    solution[1].SetODEpoint(1 * step, vec[1]);

    for (int i = 2; i < n + 1; ++i)
    {
        for (int j = 0; j < Ndim; ++j)
        {
            vec[i][j] = vec[i - 2][j] + 2 * step * f[j](solution[i - 1]);
            solution[i].SetODEpoint(i * step, vec[i]);
        }
    }
    MS.insert({"LeapFrog", solution}); //Store the solution
    //Clean
    solution.clear();
    vec.clear();
    return MS.find("LeapFrog")->second;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

const std::vector<ODEpoint> &ODESolver::EulerSolver(double step, double Time) //Using the Euler or Euler-Cromer method
{
    int n = floor(Time / step);
    //Temporary vector of ODEpoints to store the solution
    vector<ODEpoint> solution;
    solution.resize(n + 1);
    ODEpoint aux;
    //Temporary vector of vectors to store the values of theta and omega, initializing it with 0
    vector<vector<double>> vec;
    vec.resize(n + 1);
    for (int j = 0; j < n + 1; ++j)
    {
        vec[j].resize(Ndim);
    }
    //Initial conditions
    for (int j = 0; j < Ndim; ++j)
    {
        vec[0][j] = X0[j];
    }
    //Euler-Cromer
    if (onequation)
    {
        for (int i = 0; i < n; ++i)
        {
            solution[i].SetODEpoint(i * step, vec[i]); //Store solution for calculation
            aux.SetODEpoint(i * step, vec[i]);
            for (int j = Ndim - 1; j > -1; --j)
            {
                vec[i + 1][j] = vec[i][j] + step * f[j](aux);
                vec[i][j] = vec[i + 1][j];
                aux.SetODEpoint(i * step, vec[i]);
            }
        }
        solution[n].SetODEpoint(n * step, vec[n]); //Store the last solution
    }

    //Just Regular Euler
    else
    {
        for (int i = 0; i < n; ++i)
        {
            solution[i].SetODEpoint(i * step, vec[i]); //Store solution for calculation
            for (int j = 0; j < Ndim; ++j)
            {
                vec[i + 1][j] = vec[i][j] + step * f[j](solution[i]); //Methods
            }
        }
        solution[n].SetODEpoint(n * step, vec[n]); //Store the last solution
    }
    MS.insert({"Euler", solution}); //Store the solution
    //Clean
    solution.clear();
    vec.clear();
    return MS.find("Euler")->second; //Return the solution
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

const std::vector<ODEpoint> &ODESolver::TrapezoidalSolver(double step, double Time) //Using the Trapezoidal Method
{
    int n = floor(Time / step);
    //Temporary vector of ODEpoints to store the solution
    vector<ODEpoint> solution;
    solution.resize(n + 1);
    //Temporary vector of vectors to store the values of theta and omega, initializing it with 0
    vector<vector<double>> vec;
    vec.resize(n + 1);
    for (int j = 0; j < n + 1; ++j)
    {
        vec[j].resize(Ndim);
    }

    //Initial conditions for the Nondimensionalized conditions
    for (int j = 0; j < Ndim; j++)
        vec[0][j] = X0[j];

    for (int i = 0; i < n; ++i)
    {
        for (int k = 0; k < Ndim; ++k)
        {
            solution[i].SetODEpoint(i * step, vec[i]); //Temporary solution for calculation
            //Calculate the intermediary solution
            vec[i + 1][k] = vec[i][k] + step * f[k](solution[i]);
            solution[i + 1].SetODEpoint((i + 1) * step, vec[i + 1]);
            //Calculate the final values
            vec[i + 1][k] = vec[i][k] + (step / 2) * (f[k](solution[i]) + f[k](solution[i + 1]));
        }
        solution[i].SetODEpoint(i * step, vec[i]); //Store the final solution, considering the nondimensionalization
    }

    solution[n].SetODEpoint(n * step, vec[n]); //Store the last solution, considering the nondimensionalization

    MS.insert({"Trapezoidal", solution}); //Store the solution
    //Clean
    solution.clear();
    vec.clear();
    return MS.find("Trapezoidal")->second; //Return the solution
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
const std::vector<ODEpoint> &ODESolver::RungeKutta2(double step, double Time)
{
    int n = floor(Time / step);
    //Temporary vector of ODEpoints to store the solution
    vector<ODEpoint> solution;
    solution.resize(n + 1);
    //Temporary vector of vectors to store the values of theta and omega, initializing it with 0
    vector<vector<double>> vec;
    vec.resize(n + 1);
    for (int j = 0; j < n + 1; ++j)
    {
        vec[j].resize(Ndim);
    }
    //Initial conditions for the Nondimensionalized conditions
    for (int j = 0; j < Ndim; j++)
    {
        vec[0][j] = X0[j];
    }
    vector<double> K1, K2;
    K1.resize(Ndim);
    K2.resize(Ndim);
    double hep1, help2;
    for (int i = 0; i < n; ++i)
    {
        solution[i].SetODEpoint(i * step, vec[i]);
        for (int k = 0; k < Ndim; ++k)
        {
            K1[k] = step * f[k](solution[i]);
            vec[i + 1][k] = vec[i][k] + K1[k] / 2;
        }
        solution[i + 1].SetODEpoint((i + 1 / 2) * step, vec[i + 1]);
        for (int k = 0; k < Ndim; ++k)
        {
            K2[k] = step * f[k](solution[i + 1]);
            vec[i + 1][k] = vec[i][k] + K2[k];
        }
    }
    solution[n].SetODEpoint(n * step, vec[n]);
    vec.clear();
    K1.clear();
    K2.clear();

    MS.insert({"Runge-Kutta2", solution}); //Store the solution
    //Clean
    solution.clear();

    return MS.find("Runge-Kutta2")->second; //Return the solution
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

const std::vector<ODEpoint> &ODESolver::RungeKutta4(double step, double Time) //Using the RK4 method
{
    int n = floor(Time / step);
    //Temporary vector of ODEpoints to store the solution
    vector<ODEpoint> solution;
    solution.resize(n + 1);
    //Temporary vector of vectors to store the values of theta and omega, initializing it with 0
    vector<vector<double>> vec;
    vec.resize(n + 1);
    for (int j = 0; j < n + 1; ++j)
    {
        vec[j].resize(Ndim);
    }
    //Initial conditions for the Nondimensionalized conditions
    for (int j = 0; j < Ndim; j++)
        vec[0][j] = X0[j];

    vector<vector<double>> K; // Store the Matrix of K values
    K.resize(4);
    for (int j = 0; j < 4; ++j)
    {
        K[j].resize(Ndim);
    }

    //auxiliar variables
    double iter;
    int j;
    int k;
    for (int i = 0; i < n; ++i)
    {
        solution[i].SetODEpoint(i * step, vec[i]); //Temporary solution

        for (j = 0; j < 3; j++) //Calculates K1, K2, K3
        {
            iter = floor((j + 2) / 2);
            for (k = 0; k < Ndim; k++)
            {
                K[j][k] = f[k](solution[i]);
                //Use K to calculate the intermedeary next iteration of vec
                vec[i + 1][k] = vec[i][k] + step * K[j][k] * iter / 2;
            }
            solution[i].SetODEpoint((i + iter / 2) * step, vec[i + 1]); //Temporary solution
        }
        for (k = 0; k < Ndim; k++) //Calculate K4
        {
            K[3][k] = f[k](solution[i]);
            //Calculate the final value for the next iteration of vec
            vec[i + 1][k] = vec[i][k] + step / 6 * (K[0][k] + 2 * K[1][k] + 2 * K[2][k] + K[3][k]);
        }
        solution[i].SetODEpoint(i * step, vec[i]);
    }

    solution[n].SetODEpoint(n * step, vec[n]); //Store the last solution, considering the nondimensionalization

    //Clean
    vec.clear();
    K.clear();

    MS.insert({"Runge-Kutta4", solution}); //Store the solution
    //Clean
    solution.clear();

    return MS.find("Runge-Kutta4")->second; //Return the solution
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void ODESolver::SetXvar(const Xvar &newInitialCond)
{
    X0 = newInitialCond;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

const std::vector<ODEpoint> &ODESolver::Dimensionalize(std::string s)
{
    vector<ODEpoint> P; //Draw Helper

    if (MS.find(s) != MS.end()) //See if the solution is already calculated
    {
        P = MS.find(s)->second; //If yes, store it
    }

    else //If not, send a warning
    {
        cout << "Solution not calculated" << endl; //In case the string is incorrect
    }
    int nsteps = P.size();
    for (int k = 0; k < nsteps; k++)
    {
        P[k].T() *= factort;
        for (int i = 0; i < Ndim; i++)
        {
            P[k][i] *= factor[i];
        }
    }

    MDS.insert({s, P}); //Store the solution
    //Clean
    P.clear();
    return MDS.find(s)->second; //Return the solution
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void ODESolver::Draw(string s, double ti, double tf, bool PhaseDiagram, bool Dimensionalized) //Draw the solution
{
    TCanvas *c1 = new TCanvas("c1", "ODE Solver", 100, 100, 900, 500); //Create Canvas
    vector<ODEpoint> P;                                                //Draw Helper
    if (Dimensionalized)
    {
        if (MDS.find(s) != MDS.end()) //See if the solution is already calculated
        {
            P = MDS.find(s)->second; //If yes, store it
        }
        else //If not, calculate it
        {
            if (s == "Euler")
            {
                EulerSolver(1E-2, tf);
                P = Dimensionalize(s);
            }
            else if (s == "Verlet")
            {
                VerletSolver(1E-2, tf);
                P = Dimensionalize(s);
            }
            else if (s == "Trapezoidal")
            {
                TrapezoidalSolver(1E-2, tf);
                P = Dimensionalize(s);
            }
            else if (s == "Runge-Kutta4")
            {
                RungeKutta4(1E-2, tf);
                P = Dimensionalize(s);
            }
            else if (s == "LeapFrog")
            {
                LeapFrog(1E-2, tf);
                P = Dimensionalize(s);
            }
            else
            {
                cout << "The string is not valid, try: Runge-Kutta, Trapezoidal, Verlet, Euler, LeapFrog" << endl; //In case the string is incorrect
            }
        }
    }
    else
    {

        if (MS.find(s) != MS.end()) //See if the solution is already calculated
        {
            P = MS.find(s)->second; //If yes, store it
        }

        else //If not, calculate it
        {
            if (s == "Euler")
            {
                P = EulerSolver(1E-1, tf);
            }
            else if (s == "Verlet")
            {
                P = VerletSolver(1E-1, tf);
            }
            else if (s == "Trapezoidal")
            {
                P = TrapezoidalSolver(1E-1, tf);
            }
            else if (s == "Runge-Kutta")
            {
                P = RungeKutta4(1E-1, tf);
            }
            else if (s == "LeapFrog")
            {
                P = LeapFrog(1E-1, tf);
            }
            else
            {
                cout << "The string is not valid, try: Runge-Kutta, Trapezoidal, Verlet, Euler, LeapFrog" << endl; //In case the string is incorrect
            }
        }
    }

    int N = P.size(); //Get size

    double *tempo = new double[N]; //Create an array to store time

    for (int i = 0; i < N; ++i)
    {
        tempo[i] = P[i].T();
    }

    for (int j = 0; j < Ndim; ++j)
    {
        double *vars = new double[N];
        for (int i = 0; i < N; ++i)
        {
            vars[i] = P[i][j];
        }
        TGraph *gr = new TGraph(N, tempo, vars);
        gr->SetMarkerColor(kBlue);
        gr->SetLineColor(kBlue);
        std::string title = "Equation ";
        title = title + (j + 1) + "; x; y";
        gr->SetTitle(title.c_str());
        gr->SetMarkerStyle(kFullSquare);
        gr->Draw("ACP");
        if (j == 0)
        {
            c1->Print("Plots.pdf(");
        }
        else if (j == Ndim - 1)
        {
            c1->Print("Plots.pdf)");
        }
        else
        {
            c1->Print("Plots.pdf");
        }
        delete gr, vars;
    }
    delete c1;
    if (Ndim == 2)
    {
        TCanvas *c2 = new TCanvas("c2", "Graph", 100, 100, 900, 500);
        double *y = new double[N];
        double *z = new double[N];
        for (int i = 0; i < N; ++i)
        {
            y[i] = P[i][0];
            z[i] = P[i][1];
        }
        TGraph2D *G2D = new TGraph2D(N, y, z, tempo);
        G2D->Draw();
        c2->Update();
        c2->Modified();
        c2->SaveAs("3DGraph.pdf");
        c2->Clear();
        TGraph *G1D = new TGraph(N, y, z);
        G1D->Draw();
        c2->Update();
        c2->Modified();
        c2->SaveAs("PhaseDiagram.pdf");
        c2->Clear();
        delete c2, G1D, G2D;
    }
    P.clear();
}
