double g=9.81; //Global Variable gravity
#include "pendulum.h"
#include <sstream>
using namespace std;

//Switch between Cromer-Euler and Euler
#define __CROMER__

pendulum::pendulum() : L(1), X0({M_PI / 6, -1}){ Init(); } //Standard Constructor

pendulum::pendulum(double length, const Xvar &x) : L(length), X0(x) { Init(); } //XVar constructor

pendulum::pendulum(double length, const std::initializer_list<double> &v) : L(length), X0(v) { Init(); } //Initializer list constructor

pendulum::pendulum(double length, double theta_0, double theta_vel_0) : L(length), X0({theta_0, theta_vel_0}) { Init(); }//Initial Conditions as Doubles 

void pendulum::Init() //Create the lambda functions to solve the differential equation
{
   f[0] = [](ODEpoint &x)
   { return x[1]; };
   f[1] = [this](ODEpoint &x)
   { return -sin((this->X0[0]) * x[0]); };
}

const std::vector<ODEpoint> &pendulum::VerletSolver(double step, double Time) //Using the Verlet Method
{
   //Variables for the Nondimensionalization
   double fatort = sqrt((X0[0] * L) / g);
   double fatorw = sqrt((X0[0] * g) / L);
   int n = floor(Time / (step));
   double step2 = step / fatort;

   //Temporary vector of ODEpoints to store the solution
   vector<ODEpoint> solution;
   solution.resize(n + 1);

   //Temporary vector of vectors to store the values of theta and omega, initializing it with 0
   vector<vector<double>> vec;
   for (int j = 0; j < n + 1; ++j)
   {
      vec.push_back({0, 0});
   }

   //Initial conditions for the Nondimensionalized conditions 
   vec[0][0] = X0[0] / X0[0];
   vec[0][1] = X0[1] / fatorw;
   solution[0].SetODEpoint(0, {X0[0], X0[1]}); //Store it at the solution vector
   vec[1][0] = X0[0] / X0[0] - (step2 * X0[1]) / fatorw + ((step2 * step2) / 2) * (f[1](solution[0]));//Calculate the first iteration for theta

   for (int i = 1; i < n; ++i)
   {
      solution[i].SetODEpoint(i * step2, {vec[i][0], vec[i][1]}); //Temporarily create a solution to use in the next calculations
      vec[i + 1][0] = 2 * vec[i][0] - vec[i - 1][0] + pow(step2, 2) * (f[1](solution[i])); //Calculate the next theta
      vec[i][1] = (vec[i + 1][0] - vec[i - 1][0]) / (2 * step2); //Calculate omega 
      solution[i].SetODEpoint(i * fatort * step2, {X0[0] * vec[i][0], fatorw * vec[i][1]});//Store the real solution, considering the nondimensionalization
   }

   double last_theta = 2 * vec[n][0] - vec[n - 1][0] + step2 * step2 * (-sin(vec[n][0])); //Calculate an extra theta to get the last omega
   vec[n][1] = (last_theta - vec[n - 1][0]) / (2 * step2); //Last omega
   solution[n].SetODEpoint(n * fatort * step2, {X0[0] * vec[n][0], fatorw * vec[n][1]});//Store last solution, considering the nondimensionalization

   MS.insert({"Verlet", solution}); //Store the solution
   //Clean
   solution.clear();
   vec.clear();
   return MS.find("Verlet")->second; //Return the solution
}

const std::vector<ODEpoint> &pendulum::EulerSolver(double step, double Time) //Using the Euler or Euler-Cromer method
{
   //Variables for the Nondimensionalization
   double fatort = sqrt((X0[0] * L) / g);
   double fatorw = sqrt((X0[0] * g) / L);
   int n = floor(Time / step);
   double step2 = step / fatort;
   //Temporary vector of ODEpoints to store the solution
   vector<ODEpoint> solution;
   solution.resize(n + 1);
   //Temporary vector of vectors to store the values of theta and omega, initializing it with 0
   vector<vector<double>> vec;
   for (int j = 0; j < n + 1; ++j)
   {
      vec.push_back({0, 0});
   }
   //Initial conditions for the Nondimensionalized conditions 
   vec[0][0] = X0[0] / X0[0];
   vec[0][1] = X0[1] / fatorw;

#ifndef __CROMER__ //Use if we only want Euler's method
   for (int i = 0; i < n; ++i)
   {
      solution[i].SetODEpoint(i * step2, {vec[i][0], vec[i][1]}); //Store temporary solution for calculation
      vec[i + 1][0] = vec[i][0] + step2 * f[0](solution[i]);      //Methods
      vec[i + 1][1] = vec[i][1] + step2 * f[1](solution[i]);
      solution[i].SetODEpoint(i * step2 * fatort, {X0[0] * vec[i][0], fatorw * vec[i][1]}); //Store the final solution, considering the nondimensionalization
   }
#endif

#ifdef __CROMER__  //Use if we only want Euler-Cromer's method
   for (int i = 0; i < n; ++i)
   {
      solution[i].SetODEpoint(i * step2, {vec[i][0], vec[i][1]}); //Store temporary solution for calculation
      vec[i + 1][1] = vec[i][1] + step2 * f[1](solution[i]); //Calculate omega for the next iteration
      solution[i + 1].SetODEpoint((i + 1) * step2, {vec[i][0], vec[i + 1][1]}); //Store temporary solution for calculation
      vec[i + 1][0] = vec[i][0] + step2 * f[0](solution[i + 1]); //Calculat theta for the next iteration
      solution[i].SetODEpoint(i * step2 * fatort, {X0[0] * vec[i][0], fatorw * vec[i][1]}); //Store the final solution, considering the nondimensionalization
   }
#endif
   solution[n].SetODEpoint(n * step2 * fatort, {X0[0] * vec[n][0], fatorw * vec[n][1]}); //Store the last solution

   MS.insert({"Euler", solution}); //Store the solution
   //Clean
   solution.clear();
   vec.clear();
   return MS.find("Euler")->second; //Return the solution
}

const std::vector<ODEpoint> &pendulum::TrapezoidalSolver(double step, double Time)//Using the Trapezoidal Method
{
   //Variables for the Nondimensionalization
   double fatort = sqrt((X0[0] * L) / g);
   double fatorw = sqrt((X0[0] * g) / L);
   int n = floor(Time / step);
   double step2 = step / fatort;
   //Temporary vector of ODEpoints to store the solution
   vector<ODEpoint> solution;
   solution.resize(n + 1);
   //Temporary vector of vectors to store the values of theta and omega, initializing it with 0
   vector<vector<double>> vec;
   for (int j = 0; j < n + 1; ++j)
   {
      vec.push_back({0, 0});
   }
   //Initial conditions for the Nondimensionalized conditions 
   vec[0][0] = X0[0] / X0[0];
   vec[0][1] = X0[1] / fatorw;

   for (int i = 0; i < n; ++i)
   {
      solution[i].SetODEpoint(i * step2, {vec[i][0], vec[i][1]}); //Temporary solution for calculation
      //Calculate the intermediary solution
      vec[i + 1][0] = vec[i][0] + step2 * f[0](solution[i]); 
      vec[i + 1][1] = vec[i][1] + step2 * f[1](solution[i]);
      solution[i + 1].SetODEpoint((i + 1) * step2, {vec[i + 1][0], vec[i + 1][1]});
      //Calculate the final values 
      vec[i + 1][0] = vec[i][0] + step2 / 2 * (f[0](solution[i]) + f[0](solution[i + 1]));
      vec[i + 1][1] = vec[i][1] + step2 / 2 * (f[1](solution[i]) + f[1](solution[i + 1]));
      solution[i].SetODEpoint(i * step2 * fatort, {X0[0] * vec[i][0], fatorw * vec[i][1]}); //Store the final solution, considering the nondimensionalization
   }

   solution[n].SetODEpoint(n * step2 * fatort, {X0[0] * vec[n][0], fatorw * vec[n][1]}); //Store the last solution, considering the nondimensionalization

   MS.insert({"Trapezoidal", solution}); //Store the solution
   //Clean
   solution.clear();
   vec.clear();
   return MS.find("Trapezoidal")->second; //Return the solution
}

const std::vector<ODEpoint> &pendulum::RungeKutta4(double step, double Time)//Using the RK4 method
{
   //Variables for the Nondimensionalization
   double fatort = sqrt((X0[0] * L) / g);
   double fatorw = sqrt((X0[0] * g) / L);
   int n = floor(Time / step);
   double step2 = step / fatort;
   //Temporary vector of ODEpoints to store the solution
   vector<ODEpoint> solution;
   solution.resize(n + 1);
   //Temporary vector of vectors to store the values of theta and omega, initializing it with 0
   vector<vector<double>> vec;
   for (int j = 0; j < n + 1; ++j)
   {
      vec.push_back({0, 0});
   }
   //Initial conditions for the Nondimensionalized conditions 
   vec[0][0] = X0[0] / X0[0];
   vec[0][1] = X0[1] / fatorw;

   double K[4][2]; // Store the Matrix of K values
   //auxiliar variables
   double iter;
   int j; 
   for (int i = 0; i < n; ++i)
   {
      solution[i].SetODEpoint(i * step2, {vec[i][0], vec[i][1]}); //Temporary solution

      for (j = 0; j < 3; j++) //Calculates K11, K12, K21, K22, K31, K32
      {
         iter = floor((j + 2) / 2);
         K[j][0] = f[0](solution[i]); 
         K[j][1] = f[1](solution[i]); 
         //Use K to calculate the intermedeary next iteration of vec 
         vec[i + 1][0] = vec[i][0] + step2 * K[j][0] * iter / 2;
         vec[i + 1][1] = vec[i][1] + step2 * K[j][1] * iter / 2;
         solution[i].SetODEpoint((i + iter / 2) * step2, {vec[i + 1][0], vec[i + 1][1]}); //Temporary solution
      }
      K[3][0] = f[0](solution[i]); //Calculate K41
      K[3][1] = f[1](solution[i]); //Calculate K42
      //Calculate the final value for the next iteration of vec
      vec[i + 1][0] = vec[i][0] + step2 / 6 * (K[0][0] + 2 * K[1][0] + 2 * K[2][0] + K[3][0]);
      vec[i + 1][1] = vec[i][1] + step2 / 6 * (K[0][1] + 2 * K[1][1] + 2 * K[2][1] + K[3][1]);
      solution[i].SetODEpoint(i * step2 * fatort, {X0[0] * vec[i][0], fatorw * vec[i][1]}); //Store the final solution, considering the nondimensionalization
   }

   solution[n].SetODEpoint(n * step2 * fatort, {X0[0] * vec[n][0], fatorw * vec[n][1]}); //Store the last solution, considering the nondimensionalization

   MS.insert({"Runge-Kutta", solution}); //Store the solution
   //Clean
   solution.clear();
   vec.clear();
   return MS.find("Runge-Kutta")->second; //Return the solution
}

void pendulum::Draw(string s, double ti, double tf)//Draw the solution
{
   TCanvas *c1 = new TCanvas("c1", "Pendulum Solver", 100, 100, 900, 500); //Create Canvas
   vector<ODEpoint> P; //Draw Helper

   if (MS.find(s) != MS.end())//See if the solution is already calculated
   {  
      P = MS.find(s)->second; //If yes, store it
   }
   else//If no, calculate it
   {

      if (s == "Euler")
      {
         P = EulerSolver(1E-2, tf);
      }
      else if (s == "Verlet")
      {
         P = VerletSolver(1E-2, tf);
      }
      else if (s == "Trapezoidal")
      {
         P = TrapezoidalSolver(1E-2, tf);
      }
      else if (s == "Runge-Kutta")
      {
         P = RungeKutta4(1E-2, tf);
      }
      else
      {
         cout << "The string is not valid, try: Runge-Kutta, Trapezoidal, Verlet, Euler" << endl; //In case the string is incorrect
      }
   }
   
   int N = P.size();//Get size
   double tetaM=0;//Auxiliar variable to get the max theta
   double *tempo = new double[N]; //Create an array to store time
   double *teta = new double[N]; //Create an array to store theta
   double *velocidade = new double[N]; //Create an array to store omega
   for (int i = 0; i < N; ++i)
   {  //Store the calues in the arrays
      tempo[i] = P[i].T();
      teta[i] = P[i][0];
      velocidade[i] = P[i][1];
      if(teta[i]>tetaM)
      {
         tetaM=teta[i];//Get theta max
      }

   }

   //StringStream to write theta0 and omega0 in the title
   ostringstream streamT0;
   streamT0 << X0[0];
   string teta0= streamT0.str();
   ostringstream streamW0;
   streamW0 << X0[1];
   string omega0= streamW0.str();

   //Create the Graphs for the set of points
   TGraph *gr1 = new TGraph(N, tempo, teta);
   TGraph *gr2 = new TGraph(N, tempo, velocidade);
   TLegend *legend = new TLegend(0.45, 0.8, 0.55, 0.9);//Create label
   TMultiGraph* MG=new TMultiGraph(); //Create MultiGraph
   //Graph settings
   gr1->SetMarkerColor(kBlue);
   gr1->SetLineColor(kBlue);
   gr1->SetMarkerStyle(kFullSquare);
   gr2->SetMarkerColor(kRed);
   gr2->SetLineColor(kRed);
   gr2->SetMarkerStyle(kFullTriangleUp);
   MG->Add(gr1);
   MG->Add(gr2);
   //Create Title
   string title=s;
   title=title + " Method to solve a Pendulum with #theta_{0}=" + teta0 + " [rad] and #omega_{0}=" +omega0+" [rad/sec]";
   MG->SetTitle(title.c_str());
   //Axis settings
   MG->GetXaxis()->SetLimits(ti, tf);
   MG->GetXaxis()->SetTitle("Time [sec]");
   MG->GetYaxis()->SetTitle("Angle(#theta) [rad]||#omega [rad/sec]");
   MG->SetMaximum(1.5*tetaM*sqrt(g/L));
   MG->Draw("ACP");//Draw the axis, the points with a smooth curve
   //Settings for the label
   legend->SetHeader("Label", "C");
   legend->AddEntry(gr1, "Solution for #theta");
   legend->AddEntry(gr2, "Solution for #omega");
   legend->Draw();
   //Name of the file
   string name = s;
   name += "PendulumSolver.pdf";
   c1->Modified();
   c1->Update();
   c1->SaveAs(name.c_str());//Save the graph
   
   //clean
   P.clear();
   delete c1, gr1, gr2, legend, MG;
   delete[] tempo, velocidade, teta;
}
